@@ -3,7 +3,7 @@
  *
  * This file is part of fizmo.
  *
- * Copyright (c) 2009-2014 Christoph Ender.
+ * Copyright (c) 2009-2015 Christoph Ender.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -69,7 +69,7 @@
 
 #include "../locales/fizmo_ncursesw_locales.h"
 
-#define FIZMO_NCURSESW_VERSION "0.7.8"
+#define FIZMO_NCURSESW_VERSION "0.7.9"
 
 #ifdef ENABLE_X11_IMAGES
 #include <drilbo/drilbo.h>
@@ -668,14 +668,6 @@ static void print_startup_syntax()
   streams_latin1_output("\n");
 #endif //ENABLE_X11_IMAGES
 
-  /*
-  streams_latin1_output( " -s8, --force-8bit-sound: ");
-  i18n_translate(
-      fizmo_ncursesw_module_name,
-      i18n_ncursesw_FORCE_8BIT_SOUND);
-  streams_latin1_output("\n");
-  */
-
   streams_latin1_output( " -ds, --disable-sound: ");
   i18n_translate(
       fizmo_ncursesw_module_name,
