@@ -3,7 +3,7 @@
  *
  * This file is part of fizmo.
  *
- * Copyright (c) 2009-2014 Christoph Ender.
+ * Copyright (c) 2009-2015 Christoph Ender.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -110,6 +110,12 @@ static bool config_files_were_parsed = false;
 /*@null@*/ BLOCKBUF *upper_window_buffer = NULL;
 #endif // DISABLE_BLOCKBUFFER
 
+#ifndef DISABLE_OUTPUT_HISTORY
+void (*paragraph_attribute_function)(int *parameter1, int *parameter2) = NULL;
+void (*paragraph_removal_function)(int parameter1, int parameter2) = NULL;
+#endif // DISABLE_OUTPUT_HISTORY
+
+
 
 // "load_z_story" returns malloc()ed z_story, may be freed using free_z_story().
 static struct z_story *load_z_story(z_file *story_stream, z_file *blorb_stream)
@@ -1141,6 +1147,19 @@ char *unquote_special_chars(char *s)
 }
 
 
+#ifndef DISABLE_OUTPUT_HISTORY
+void fizmo_register_paragraph_attribute_function(
+    void (*new_paragraph_attribute_function)(int *parameter1,
+      int *parameter2)) {
+  paragraph_attribute_function = new_paragraph_attribute_function;
+}
+void fizmo_register_paragraph_removal_function(
+    void (*new_paragraph_removal_function)(int parameter1, int parameter2)) {
+  paragraph_removal_function = new_paragraph_removal_function;
+}
+#endif // DISABLE_OUTPUT_HISTORY
+
+
 void fizmo_start(z_file* story_stream, z_file *blorb_stream,
     z_file *restore_on_start_file)
 {
