@@ -1,9 +1,9 @@
 
 /* history.c
- *
+
  * This file is part of fizmo.
  *
- * Copyright (c) 2009-2014 Christoph Ender.
+ * Copyright (c) 2009-2015 Christoph Ender.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -16,7 +16,7 @@
  *    documentation and/or other materials provided with the distribution.
  * 3. The name of the author may not be used to endorse or promote products
  *    derived from this software without specific prior written permission.
- * 
+ *
  * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
@@ -70,7 +70,7 @@
  */
 
 
-#ifndef history_c_INCLUDED 
+#ifndef history_c_INCLUDED
 #define history_c_INCLUDED
 
 #include <stdlib.h>
@@ -117,6 +117,7 @@ OUTPUTHISTORY *create_outputhistory(int window_number,
   result->z_history_buffer_back_index = NULL;
   result->nof_wraparounds = 0;
   result->last_metadata_block_index = 0;
+  result->next_newline_after_buffer_back = NULL;
 
   result->history_buffer_back_index_font = font;
   result->history_buffer_back_index_style = style;
@@ -154,15 +155,21 @@ static size_t get_buffer_space_used(OUTPUTHISTORY *h)
 static void process_buffer_back(OUTPUTHISTORY *h, long nof_zucs_chars)
 {
   z_ucs *current_index = h->z_history_buffer_front_index;
+  z_ucs buf1, buf2;
 
   TRACE_LOG("Advancing buffer end from %p by %ld chars.\n",
       current_index,
       (long int)nof_zucs_chars);
 
-  do
-  {
+  do {
     TRACE_LOG("current-index: %p.\n", current_index);
 
+    // Check if we have caught up to the last processed paragraph-attribute
+    // position.
+    if (h->next_newline_after_buffer_back == current_index) {
+      h->next_newline_after_buffer_back = NULL;
+    }
+
     if (*current_index == HISTORY_METADATA_ESCAPE)
     {
       // We've found an escape code. Advance index to escape code type.
@@ -221,6 +228,30 @@ static void process_buffer_back(OUTPUTHISTORY *h, long nof_zucs_chars)
         h->history_buffer_back_index_background
           = (z_colour)*current_index;
       }
+      else if (*current_index == HISTORY_METADATA_TYPE_PARAGRAPHATTRIBUTE) {
+        // do nothing but catch the case so we're not running into the
+        // error-else below.
+        current_index
+          = current_index == h->z_history_buffer_end
+          ? h->z_history_buffer_start
+          : current_index + 1;
+        nof_zucs_chars--;
+        buf1 = *current_index;
+
+        // Advance to second parameter
+        current_index
+          = current_index == h->z_history_buffer_end
+          ? h->z_history_buffer_start
+          : current_index + 1;
+        nof_zucs_chars--;
+
+        if ( (paragraph_removal_function != NULL)
+            && (h->next_newline_after_buffer_back == NULL) ) {
+          paragraph_removal_function(
+              buf1 - HISTORY_METADATA_DATA_OFFSET,
+              *current_index - HISTORY_METADATA_DATA_OFFSET);
+        }
+      }
       else
       {
         TRACE_LOG("Inconsistent history metadata.\n");
@@ -245,6 +276,77 @@ static void process_buffer_back(OUTPUTHISTORY *h, long nof_zucs_chars)
   // history bugger is specified to have minimum buffer size of a complete
   // metadata entry though, this will work without problems.
 
+  if ( (paragraph_removal_function != NULL)
+      && (h->next_newline_after_buffer_back == NULL) ) {
+    //printf("passed next_newline_after_buffer_back.\n");
+    // Either we have never run "process_buffer_back" before, or we have
+    // passed the last evaluated paragraph-attribute position in the loop
+    // above. Either way, we now have to look for the next paragraph in
+    // order to keep the to keep a consistent paragraph attribute situation
+    // (meaning: Only whole paragraphs in the history count).
+
+    h->next_newline_after_buffer_back = current_index;
+
+    while (*current_index != Z_UCS_NEWLINE) {
+
+      if (current_index == h->z_history_buffer_front_index) {
+        // Only a single block of text in the buffer? Well, okay.
+        //printf("XXXX\n");
+        break;
+      }
+
+      // Advance 1 char.
+      current_index
+        = current_index == h->z_history_buffer_end
+        ? h->z_history_buffer_start
+        : current_index + 1;
+
+      // In case there's some metadata at this point, process it.
+      if (*current_index == HISTORY_METADATA_ESCAPE) {
+
+        // Advance to metadata type.
+        current_index
+          = current_index == h->z_history_buffer_end
+          ? h->z_history_buffer_start
+          : current_index + 1;
+        // Remember this in buf1.
+        buf1 = *current_index;
+
+        // Advance to first parameter.
+        current_index
+          = current_index == h->z_history_buffer_end
+          ? h->z_history_buffer_start
+          : current_index + 1;
+        // Remember this in buf2.
+        buf2 = *current_index;
+
+        if ( (buf1 == HISTORY_METADATA_TYPE_PARAGRAPHATTRIBUTE)
+            || (buf1 == HISTORY_METADATA_TYPE_COLOUR) ) {
+
+          // Advance to second parameter.
+          current_index
+            = current_index == h->z_history_buffer_end
+            ? h->z_history_buffer_start
+            : current_index + 1;
+          nof_zucs_chars--;
+
+          if (buf1 == HISTORY_METADATA_TYPE_PARAGRAPHATTRIBUTE) {
+            paragraph_removal_function(
+                buf2 - HISTORY_METADATA_DATA_OFFSET,
+                *current_index - HISTORY_METADATA_DATA_OFFSET);
+          }
+        }
+      }
+    }
+
+    //printf("bufferback: next_newline_after_buffer_back %p.\n",
+    //    h->next_newline_after_buffer_back);
+
+    h->next_newline_after_buffer_back = current_index;
+  }
+
+  //result->next_newline_after_buffer_back = NULL;
+
   TRACE_LOG("Processed chars up to before %p.\n", current_index);
 }
 
@@ -292,32 +394,26 @@ static void try_to_enlarge_buffer(OUTPUTHISTORY *h, size_t desired_z_ucs_size)
 }
 
 
+// This method helps to ensure that we have at least one entry for each
+// metadata type which is required to display on the screen (style, font
+// and color) every Z_HISTORY_METADATA_STATE_BLOCK_SIZE bytes. In case
+// we wouldn't be doing this, it would mean that games which never write a
+// certain metadata type (games without color support would for example never
+// write color metadata) we'd have to walk back the entire buffer to be able
+// to evaluate the color of the current char we're pointing at.
 static void write_metadata_state_block_if_necessary(OUTPUTHISTORY *h)
 {
   long int metadata_block_index;
   long int buffer_index
     = h->z_history_buffer_front_index - h->z_history_buffer_start;
 
-  TRACE_LOG("xyz: %ld, %d, %ld\n",
-      buffer_index,
-      Z_HISTORY_METADATA_STATE_BLOCK_SIZE,
-      buffer_index % (Z_HISTORY_METADATA_STATE_BLOCK_SIZE));
-
   metadata_block_index
     = buffer_index - (buffer_index % (Z_HISTORY_METADATA_STATE_BLOCK_SIZE));
 
-  /*
-  metadata_block_index
-    = buffer_index > Z_HISTORY_METADATA_STATE_BLOCK_SIZE
-    ? buffer_index % Z_HISTORY_METADATA_STATE_BLOCK_SIZE
-    : 0;
-    */
-
   TRACE_LOG("block_index %ld(%ld), last block %ld.\n",
       metadata_block_index, buffer_index, h->last_metadata_block_index);
 
-  if (metadata_block_index != h->last_metadata_block_index)
-  {
+  if (metadata_block_index != h->last_metadata_block_index) {
     // We've now crossed a metadata block state boundary and thus we'll now
     // write the current state block.
 
@@ -417,7 +513,7 @@ void store_data_in_history(OUTPUTHISTORY *h, z_ucs *data, size_t len,
         = new_size > h->z_history_maximum_buffer_size
         ? h->z_history_maximum_buffer_size
         : new_size;
-      
+
       if (desired_size > h->z_history_buffer_size)
         try_to_enlarge_buffer(h, desired_size);
     }
@@ -550,9 +646,11 @@ void store_data_in_history(OUTPUTHISTORY *h, z_ucs *data, size_t len,
 void store_z_ucs_output_in_history(OUTPUTHISTORY *h, z_ucs *z_ucs_output)
 {
   size_t len;
+/*
 #ifdef ENABLE_TRACING
   z_ucs *cl;
 #endif // ENABLE_TRACING
+*/
 
   if (z_ucs_output == NULL)
     return;
@@ -562,6 +660,7 @@ void store_z_ucs_output_in_history(OUTPUTHISTORY *h, z_ucs *z_ucs_output)
 
   store_data_in_history(h, z_ucs_output, len, true);
 
+/*
 #ifdef ENABLE_TRACING
   cl = get_current_line(h);
   TRACE_LOG("currentline:|");
@@ -572,6 +671,7 @@ void store_z_ucs_output_in_history(OUTPUTHISTORY *h, z_ucs *z_ucs_output)
   }
   TRACE_LOG("|line end.\n");
 #endif // ENABLE_TRACING
+  */
 }
 
 
@@ -589,6 +689,8 @@ int store_metadata_in_history(OUTPUTHISTORY *h, int metadata_type, ...)
       (metadata_type != HISTORY_METADATA_TYPE_STYLE)
       &&
       (metadata_type != HISTORY_METADATA_TYPE_COLOUR)
+      &&
+      (metadata_type != HISTORY_METADATA_TYPE_PARAGRAPHATTRIBUTE)
      )
     return -1;
 
@@ -603,8 +705,10 @@ int store_metadata_in_history(OUTPUTHISTORY *h, int metadata_type, ...)
   // warning: ‘int16_t’ is promoted to ‘int’ when passed through ‘...’
   // (so you should pass ‘int’ not ‘int16_t’ to ‘va_arg’)
   parameter = va_arg(ap, int);
-  if ( (parameter < -2) || (parameter > 15 ) )
-  {
+
+  // TODO: Add verifications for other metadata-types.
+  if ( (metadata_type == HISTORY_METADATA_TYPE_COLOUR)
+           && ( (parameter < -2) || (parameter > 15 ) ) )   {
     // -2 is the lowest allowed value for Z_COLOUR_UNDEFINED, 15 the maximum
     // for all combinations of Z_STYLE.
     TRACE_LOG("Parameter value %d outside valid range.\n", parameter);
@@ -630,6 +734,10 @@ int store_metadata_in_history(OUTPUTHISTORY *h, int metadata_type, ...)
     h->history_buffer_front_index_foreground = parameter;
     TRACE_LOG("storing colour.\n");
   }
+  else if (metadata_type == HISTORY_METADATA_TYPE_PARAGRAPHATTRIBUTE)
+  {
+    TRACE_LOG("storing paragraph attribute.\n");
+  }
 
   // All parameter values are offset by +13. This is necessary to avoid
   // having LF characters in the buffer, which makes searching for paragraph
@@ -637,20 +745,21 @@ int store_metadata_in_history(OUTPUTHISTORY *h, int metadata_type, ...)
   output_buffer[2] = (z_ucs)(parameter + HISTORY_METADATA_DATA_OFFSET);
   TRACE_LOG("param1: %d.\n", parameter);
 
-  if (metadata_type == HISTORY_METADATA_TYPE_COLOUR)
-  {
+  if ( (metadata_type == HISTORY_METADATA_TYPE_COLOUR)
+          || (metadata_type == HISTORY_METADATA_TYPE_PARAGRAPHATTRIBUTE) ) {
+    // Read second parameter for all metadatatypes which require it.
     parameter = va_arg(ap, int);
     TRACE_LOG("param2: %d.\n", parameter);
-    if ( (parameter < -2) || (parameter > 15 ) )
-    {
+    if ( (metadata_type == HISTORY_METADATA_TYPE_COLOUR)
+          && ( (parameter < -2) || (parameter > 15 ) ) ) {
       TRACE_LOG("Parameter value %d outside valid range.\n", parameter);
       i18n_translate_and_exit(
           libfizmo_module_name,
           i18n_libfizmo_INVALID_PARAMETER_TYPE_P0S,
           -1,
           "parameter");
+      h->history_buffer_front_index_background = parameter;
     }
-    h->history_buffer_front_index_background = parameter;
     output_buffer[3] = (z_ucs)(parameter + HISTORY_METADATA_DATA_OFFSET);
     len = 4;
   }
@@ -689,28 +798,21 @@ z_ucs *decrement_buffer_pointer(OUTPUTHISTORY *h, z_ucs *ptr,
     unsigned int *nof_wraparounds)
 {
   if (
-      (ptr == h->z_history_buffer_back_index) 
-      &&
-      (ptr == h->z_history_buffer_front_index)
-      &&
-      (*nof_wraparounds > 0)
-     )
-  {
+      (ptr == h->z_history_buffer_back_index)
+      && (ptr == h->z_history_buffer_front_index)
+      && (*nof_wraparounds > 0)) {
     TRACE_LOG("History index already at buffer back.\n");
     return NULL;
   }
 
   ptr--;
 
-  if (ptr < h->z_history_buffer_start)
-  {
-    if (h->nof_wraparounds == 0)
-    {
+  if (ptr < h->z_history_buffer_start) {
+    if (h->nof_wraparounds == 0) {
       TRACE_LOG("History index at front of non-wrapped buffer.\n");
       return NULL;
     }
-    else
-    {
+    else {
       TRACE_LOG("History index at front, wrapping around.\n");
       ptr = h->z_history_buffer_end;
       (*nof_wraparounds)--;
@@ -739,7 +841,11 @@ int remove_chars_from_history(OUTPUTHISTORY *history, int nof_chars)
 
     if ( (*ptr == HISTORY_METADATA_ESCAPE) && (last_data != 0) )
     {
-      nof_chars += (last_data = HISTORY_METADATA_TYPE_COLOUR ? 4 : 3);
+      nof_chars
+        += ( (last_data == HISTORY_METADATA_TYPE_COLOUR)
+            || (last_data == HISTORY_METADATA_TYPE_PARAGRAPHATTRIBUTE) )
+        ? 4
+        : 3;
     }
     else
     {
@@ -774,6 +880,11 @@ static z_ucs* find_older_paragraph(OUTPUTHISTORY *h, z_ucs *index)
 }
 
 
+/*
+// The "get_current_line" function is deprecated. The problem is that the
+// conecpt of getting the current line/prompt in a single string cannot
+// work once metadata is encountered.
+//
 z_ucs *get_current_line(OUTPUTHISTORY *h)
 {
   z_ucs *last_stored_char_index, *last_newline_index, *result;
@@ -833,6 +944,7 @@ z_ucs *get_current_line(OUTPUTHISTORY *h)
 
   return result;
 }
+*/
 
 
 void destroy_history_output(history_output *output)
@@ -843,44 +955,66 @@ void destroy_history_output(history_output *output)
 
 // The OUTPUTHISTORY object will ony be valid as long as nothing new is
 // stored in the history while using it.
-history_output *init_history_output(OUTPUTHISTORY *h, history_output_target *t)
+history_output *init_history_output(OUTPUTHISTORY *h, history_output_target *t,
+    int output_init_flags)
 {
   history_output *result;
 
-  if ( (h == NULL) || (h->z_history_buffer_size == 0) )
+  if ( (h == NULL) || (h->z_history_buffer_size == 0) ) {
     return NULL;
+  }
 
   result = fizmo_malloc(sizeof(history_output));
 
   result->history = h;
   result->validity_wraparounds = h->nof_wraparounds;
   result->validity_frontindex = h->z_history_buffer_front_index;
-  result->target=t;
-  result->current_paragraph_index = h->z_history_buffer_front_index;
-  result->font_at_index = h->history_buffer_front_index_font;
-  result->style_at_index = h->history_buffer_front_index_style;
-  result->foreground_at_index = h->history_buffer_front_index_foreground;
-  result->background_at_index = h->history_buffer_front_index_background;
-  result->nof_wraparounds = 0;
-  result->found_end_of_buffer = false;
+  result->target = t;
   result->rewound_paragraph_was_newline_terminated = false;
-  result->first_iteration_done = false;
+  result->validation_disabled
+    = (output_init_flags & Z_HISTORY_OUTPUT_WITHOUT_VALIDATION) == 0
+    ? false : true;
   result->last_rewinded_paragraphs_block_index = -1;
   result->last_used_metadata_state_font = -1;
   result->last_used_metadata_state_style = -1;
   result->last_used_metadata_state_foreground = Z_COLOUR_UNDEFINED;
   result->last_used_metadata_state_background = Z_COLOUR_UNDEFINED;
+  result->last_paragraph_attribute_index = NULL;
+  result->dont_skip_newline = false;
 
-  // Since "z_history_buffer_front_index" always points to the place where
-  // the next char will be stored, we actually have to go back one char
-  // in order to find the last paragraph's stored char.
-  if ((result->current_paragraph_index = decrement_buffer_pointer(
-          result->history,
-          result->current_paragraph_index,
-          &result->nof_wraparounds)) == NULL)
-  {
-    free(result);
-    return NULL;
+  if ((output_init_flags & Z_HISTORY_OUTPUT_FROM_BUFFERBACK) == 0) {
+    TRACE_LOG("Init from buffer front.\n");
+    result->current_paragraph_index = h->z_history_buffer_front_index;
+    result->font_at_index = h->history_buffer_front_index_font;
+    result->style_at_index = h->history_buffer_front_index_style;
+    result->foreground_at_index = h->history_buffer_front_index_foreground;
+    result->background_at_index = h->history_buffer_front_index_background;
+    result->found_end_of_buffer = false;
+    result->nof_wraparounds = 0;
+    result->first_iteration_done = false;
+ 
+    // Since "z_history_buffer_front_index" always points to the place where
+    // the next char will be stored, we actually have to go back one char
+    // in order to find the last paragraph's stored char.
+    if ((result->current_paragraph_index = decrement_buffer_pointer(
+            result->history,
+            result->current_paragraph_index,
+            &result->nof_wraparounds)) == NULL) {
+      free(result);
+      return NULL;
+    }
+  }
+  else {
+    TRACE_LOG("Init from buffer back.\n");
+    result->current_paragraph_index = h->z_history_buffer_back_index;
+    result->font_at_index = h->history_buffer_back_index_font;
+    result->style_at_index = h->history_buffer_back_index_style;
+    result->foreground_at_index = h->history_buffer_back_index_foreground;
+    result->background_at_index = h->history_buffer_back_index_background;
+    result->found_end_of_buffer = true;
+    result->nof_wraparounds
+      = h->nof_wraparounds > 0 ? h->nof_wraparounds - 1 : 0;
+    result->first_iteration_done = true;
   }
 
   return result;
@@ -912,7 +1046,9 @@ static void evaluate_metadata_for_paragraph(history_output *output)
   z_ucs *i2 = NULL, *i3 = NULL, *i4 = NULL;
   int metadata_type, parameter;
 
-  validate_outputhistory(output);
+  if (output->validation_disabled == false) {
+    validate_outputhistory(output);
+  }
 
   TRACE_LOG("Evaluating metadata for current paragraph from %p.\n",
       output->current_paragraph_index);
@@ -986,20 +1122,22 @@ static void evaluate_metadata_for_paragraph(history_output *output)
       {
         TRACE_LOG("Hit end of buffer. Using back values to fill in.\n");
 
-        if (output->font_at_index == -1)
+        if (output->font_at_index == -1) {
           output->font_at_index = h->history_buffer_back_index_font;
+        }
 
-        if (output->style_at_index == -1)
-        {
+        if (output->style_at_index == -1) {
           output->style_at_index = h->history_buffer_back_index_style;
           TRACE_LOG("sai: #3\n");
         }
 
-        if (output->foreground_at_index == Z_COLOUR_UNDEFINED)
+        if (output->foreground_at_index == Z_COLOUR_UNDEFINED) {
           output->foreground_at_index=h->history_buffer_front_index_foreground;
+        }
 
-        if (output->background_at_index == Z_COLOUR_UNDEFINED)
+        if (output->background_at_index == Z_COLOUR_UNDEFINED) {
           output->background_at_index=h->history_buffer_front_index_background;
+        }
 
         break;
       }
@@ -1059,16 +1197,19 @@ static void evaluate_metadata_for_paragraph(history_output *output)
 // value is 0, in case the buffer back was encountered 1 and a negative value
 // in case of an error. In case char_count is non-null, the number of
 // non-metadata chars in this paragraph is stored at this reference.
-int output_rewind_paragraph(history_output *output, long *char_count)
+int output_rewind_paragraph(history_output *output, long *char_count,
+    int *paragraph_attr1, int *paragraph_attr2)
 {
-  z_ucs *index, *last_index;
+  z_ucs *index, *last_index, *last_index2, *last_index3;
   int nof_chars = 0;
   unsigned int nof_wraparounds, last_nof_wraparounds;
 
   TRACE_LOG("Rewinding output history by one paragraph from %p.\n",
       output->current_paragraph_index);
 
-  validate_outputhistory(output);
+  if (output->validation_disabled == false) {
+    validate_outputhistory(output);
+  }
 
   if ( (output == NULL)
       || (output->history == NULL)
@@ -1095,14 +1236,24 @@ int output_rewind_paragraph(history_output *output, long *char_count)
     // always newline-terminated.
     output->rewound_paragraph_was_newline_terminated = true;
 
-    if ((index = decrement_buffer_pointer(
-            output->history,
-            index,
-            &nof_wraparounds)) == NULL) {
-      TRACE_LOG("Couldn't execute initial index decrement.\n");
-      return -3;
+    // It's possible that we're already on the newline in case the very
+    // first paragraph didn't contain anything but a newline, so we'll
+    // only rewind further in case we're not yet there.
+    if (output->dont_skip_newline == false) {
+      TRACE_LOG("Rewinding from %p.\n", index);
+      if ((index = decrement_buffer_pointer(
+              output->history,
+              index,
+              &nof_wraparounds)) == NULL) {
+        TRACE_LOG("Couldn't execute initial index decrement.\n");
+        return -3;
+      }
+    }
+    else {
+      output->dont_skip_newline = false;
     }
 
+    TRACE_LOG("Looking at %p.\n", index);
     if (*index != '\n') {
       TRACE_LOG("Internal error rewinding.\n");
       return -4;
@@ -1137,6 +1288,7 @@ int output_rewind_paragraph(history_output *output, long *char_count)
   }
   else {
     if (*index == '\n') {
+      output->dont_skip_newline = true;
       TRACE_LOG("Last output char is newline, returning from 1st iteration.\n");
       output->first_iteration_done = true;
       output->metadata_at_index_evaluated = false;
@@ -1148,6 +1300,7 @@ int output_rewind_paragraph(history_output *output, long *char_count)
       return 0;
     }
     else {
+      output->dont_skip_newline = false;
       output->rewound_paragraph_was_newline_terminated = false;
     }
 
@@ -1162,11 +1315,17 @@ int output_rewind_paragraph(history_output *output, long *char_count)
   // to rewind over. We're rewinding until we find the newline or the
   // buffer start. In the latter case we've got a non-full paragraph which
   // we won't return.
+  last_index = NULL;
+  last_index2 = NULL;
   do {
     // In this loop we're remembering the last index position for two
     // reasons: One to be able to read the last z_ucs we've iterated over
-    // for eaasier metadata evaluation (see below), second to make
-    // skipping the newline we might find a bit easier.
+    // for easier metadata evaluation (see below), second to make
+    // skipping the newline we might find a bit easier. We're also
+    // remebering the second- and third-last-index for paragraph attribute
+    // envaluation.
+    last_index3 = last_index2;
+    last_index2 = last_index;
     last_index = index;
     last_nof_wraparounds = nof_wraparounds;
 
@@ -1198,7 +1357,24 @@ int output_rewind_paragraph(history_output *output, long *char_count)
     nof_chars++;
 
     if (*index == HISTORY_METADATA_ESCAPE) {
-      nof_chars -= (*last_index == HISTORY_METADATA_TYPE_COLOUR ? 4 : 3);
+      if (*last_index == HISTORY_METADATA_TYPE_COLOUR) {
+        nof_chars -= 4;
+      }
+      else if (*last_index == HISTORY_METADATA_TYPE_PARAGRAPHATTRIBUTE) {
+
+        if (paragraph_attr1 != NULL) {
+          *paragraph_attr1 = *last_index2 - HISTORY_METADATA_DATA_OFFSET;
+        }
+
+        if (paragraph_attr2 != NULL) {
+          *paragraph_attr2 = *last_index3 - HISTORY_METADATA_DATA_OFFSET;
+        }
+
+        nof_chars -= 4;
+      }
+      else {
+        nof_chars -= 3;
+      }
     }
 
     TRACE_LOG("Index pointing at '%c' / %p.\n", *index, index);
@@ -1237,9 +1413,39 @@ int output_rewind_paragraph(history_output *output, long *char_count)
 }
 
 
-bool is_output_at_frontindex(history_output *output) {
+int alter_last_paragraph_attributes(history_output *output,
+    int paragraph_attr1, int paragraph_attr2) {
+  z_ucs *index;
+
   validate_outputhistory(output);
 
+  if (output->last_paragraph_attribute_index == NULL) {
+    TRACE_LOG("Not altering paragraph attributes, pointer is NULL.\n");
+    return -1;
+  }
+
+  TRACE_LOG("Altering paragraph attributes to %d and %d.\n",
+      paragraph_attr1, paragraph_attr2);
+
+  index = output->last_paragraph_attribute_index;
+
+  *index = (z_ucs)(paragraph_attr1 + HISTORY_METADATA_DATA_OFFSET);
+  if (++index > output->history->z_history_buffer_end) {
+    index = output->history->z_history_buffer_start;
+  }
+
+  *index = (z_ucs)(paragraph_attr2 + HISTORY_METADATA_DATA_OFFSET);
+
+  return 0;
+}
+
+
+bool is_output_at_frontindex(history_output *output) {
+
+  if (output->validation_disabled == false) {
+    validate_outputhistory(output);
+  }
+
   if (output->current_paragraph_index
       == output->history->z_history_buffer_front_index) {
     return true;
@@ -1258,7 +1464,9 @@ int output_repeat_paragraphs(history_output *output, int n,
   int buf_index;
   int metadata_type = -1, parameter, parameter2;
 
-  validate_outputhistory(output);
+  if (output->validation_disabled == false) {
+    validate_outputhistory(output);
+  }
 
   if (include_metadata == true)
     evaluate_metadata_for_paragraph(output);
@@ -1286,85 +1494,106 @@ int output_repeat_paragraphs(history_output *output, int n,
     output->found_end_of_buffer = false;
 
   buf_index = 0;
-  while (n > 0)
-  {
-    TRACE_LOG("Looking at %p.\n", output_ptr);
 
-    if (output_ptr == output->history->z_history_buffer_front_index)
+  if (output_ptr == output->history->z_history_buffer_front_index) {
+    TRACE_LOG("Already at buffer front.\n");
+    n = -1;
+  }
+  else {
+    while (n > 0)
     {
-      n = -1;
-      TRACE_LOG("Buffer front encountered.\n");
-    }
-    else if (*output_ptr == '\n')
-      n--;
+      TRACE_LOG("Looking at %p.\n", output_ptr);
 
-    if ( (buf_index == REPEAT_PARAGRAPH_BUF_SIZE - 1)
-        || (n < 1)
-        || (*output_ptr == HISTORY_METADATA_ESCAPE) ) {
-      output_buf[buf_index] = 0;
-      TRACE_LOG("Sending %d char(s) of output.\n", buf_index);
-      output->target->z_ucs_output(output_buf);
-
-      if (n < 1)
-      {
-        TRACE_LOG("n < 1.\n");
-        break;
+      if (*output_ptr == '\n') {
+        n--;
       }
 
-      buf_index = 0;
+      if ( (buf_index == REPEAT_PARAGRAPH_BUF_SIZE - 1)
+          || (n < 1)
+          || (*output_ptr == HISTORY_METADATA_ESCAPE)
+          || (output_ptr == output->history->z_history_buffer_front_index) ) {
+        output_buf[buf_index] = 0;
+        TRACE_LOG("Sending %d char(s) of output.\n", buf_index);
+        output->target->z_ucs_output(output_buf);
 
-      if (*output_ptr == HISTORY_METADATA_ESCAPE)
-      {
-        TRACE_LOG("Metadata found at %p in output.\n", output_ptr);
-
-        if (++output_ptr > output->history->z_history_buffer_end)
-          output_ptr = output->history->z_history_buffer_start;
-        metadata_type = *output_ptr;
-        if (++output_ptr > output->history->z_history_buffer_end)
-          output_ptr = output->history->z_history_buffer_start;
-        parameter = (int)*output_ptr - HISTORY_METADATA_DATA_OFFSET;
-
-        if (metadata_type == HISTORY_METADATA_TYPE_FONT)
-        {
-          output->font_at_index = parameter;
-          if (include_metadata == true)
-            output->target->set_font(parameter);
+        if (output_ptr == output->history->z_history_buffer_front_index) {
+          TRACE_LOG("Buffer front encountered.\n");
+          break;
         }
-        else if (metadata_type == HISTORY_METADATA_TYPE_STYLE)
+
+        if (n < 1)
         {
-          output->style_at_index = parameter;
-          if (include_metadata == true)
-            output->target->set_text_style(parameter);
+          TRACE_LOG("n < 1.\n");
+          break;
         }
-        else if (metadata_type == HISTORY_METADATA_TYPE_COLOUR)
-        {
+
+        buf_index = 0;
+
+        if (*output_ptr == HISTORY_METADATA_ESCAPE) {
+
+          TRACE_LOG("Metadata found at %p in output.\n", output_ptr);
+
           if (++output_ptr > output->history->z_history_buffer_end)
             output_ptr = output->history->z_history_buffer_start;
-          parameter2 = (int)*output_ptr - HISTORY_METADATA_DATA_OFFSET;
-          output->foreground_at_index = parameter;
-          output->background_at_index = parameter2;
-          if (include_metadata == true)
-            output->target->set_colour(parameter, parameter2, -1);
-        }
-        else
-        {
-          TRACE_LOG("Invalid metadata type %d\n", metadata_type);
-          i18n_translate_and_exit(
-              libfizmo_module_name,
-              i18n_libfizmo_INVALID_PARAMETER_TYPE_P0S,
-              -1,
-              "metadata");
+          metadata_type = *output_ptr;
+          if (++output_ptr > output->history->z_history_buffer_end)
+            output_ptr = output->history->z_history_buffer_start;
+          parameter = (int)*output_ptr - HISTORY_METADATA_DATA_OFFSET;
+
+          if (metadata_type == HISTORY_METADATA_TYPE_FONT)
+          {
+            output->font_at_index = parameter;
+            if (include_metadata == true)
+              output->target->set_font(parameter);
+          }
+          else if (metadata_type == HISTORY_METADATA_TYPE_STYLE)
+          {
+            output->style_at_index = parameter;
+            if (include_metadata == true)
+              output->target->set_text_style(parameter);
+          }
+          else if (metadata_type == HISTORY_METADATA_TYPE_COLOUR)
+          {
+            if (++output_ptr > output->history->z_history_buffer_end)
+              output_ptr = output->history->z_history_buffer_start;
+            parameter2 = (int)*output_ptr - HISTORY_METADATA_DATA_OFFSET;
+            output->foreground_at_index = parameter;
+            output->background_at_index = parameter2;
+            if (include_metadata == true)
+              output->target->set_colour(parameter, parameter2, -1);
+          }
+          else if (metadata_type == HISTORY_METADATA_TYPE_PARAGRAPHATTRIBUTE)
+          {
+            output->last_paragraph_attribute_index = output_ptr;
+
+            // Don't do anything but catch the case in order that we're
+            // not running into the error-else below.
+            if (++output_ptr > output->history->z_history_buffer_end)
+              output_ptr = output->history->z_history_buffer_start;
+            //parameter2 = (int)*output_ptr - HISTORY_METADATA_DATA_OFFSET;
+          }
+          else
+          {
+            TRACE_LOG("Invalid metadata type %d\n", metadata_type);
+            i18n_translate_and_exit(
+                libfizmo_module_name,
+                i18n_libfizmo_INVALID_PARAMETER_TYPE_P0S,
+                -1,
+                "metadata");
+          }
         }
       }
+
+      if (metadata_type == -1) {
+        output_buf[buf_index++] = *output_ptr;
+      }
+      else {
+        metadata_type = -1;
+      }
+
+      if ((++output_ptr) > output->history->z_history_buffer_end)
+        output_ptr = output->history->z_history_buffer_start;
     }
-
-    if (metadata_type == -1)
-      output_buf[buf_index++] = *output_ptr;
-    else
-      metadata_type = -1;
-
-    if ((++output_ptr) > output->history->z_history_buffer_end)
-      output_ptr = output->history->z_history_buffer_start;
   }
 
   TRACE_LOG("n: %d.\n", n);
@@ -1409,7 +1638,9 @@ int output_repeat_paragraphs(history_output *output, int n,
 
 void remember_history_output_position(history_output *output)
 {
-  validate_outputhistory(output);
+  if (output->validation_disabled == false) {
+    validate_outputhistory(output);
+  }
 
   output->saved_current_paragraph_index =
     output->current_paragraph_index;
@@ -1446,7 +1677,9 @@ void remember_history_output_position(history_output *output)
 
 void restore_history_output_position(history_output *output)
 {
-  validate_outputhistory(output);
+  if (output->validation_disabled == false) {
+    validate_outputhistory(output);
+  }
 
   output->current_paragraph_index =
     output->saved_current_paragraph_index;
