@@ -3,7 +3,7 @@
  *
  * This file is part of fizmo.
  *
- * Copyright (c) 2009-2014 Christoph Ender.
+ * Copyright (c) 2009-2015 Christoph Ender.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -42,6 +42,7 @@
 #define HISTORY_METADATA_TYPE_FONT 1
 #define HISTORY_METADATA_TYPE_STYLE 2
 #define HISTORY_METADATA_TYPE_COLOUR 3
+#define HISTORY_METADATA_TYPE_PARAGRAPHATTRIBUTE 4
 
 #define HISTORY_METADATA_DATA_OFFSET 13
 
@@ -49,6 +50,10 @@
 #define Z_HISTORY_MAXIMUM_SIZE 128*1024*1024 // given in units of z_ucs
 #define Z_HISTORY_METADATA_STATE_BLOCK_SIZE 16*1024 // given in units of z_ucs
 
+#define Z_HISTORY_OUTPUT_WITHOUT_EXTRAS 0
+#define Z_HISTORY_OUTPUT_FROM_BUFFERBACK 1
+#define Z_HISTORY_OUTPUT_WITHOUT_VALIDATION 2
+
 
 typedef struct
 {
@@ -70,6 +75,7 @@ typedef struct
   z_ucs *z_history_buffer_back_index;
 
   long int last_metadata_block_index;
+  z_ucs *next_newline_after_buffer_back;
 
   // These four values are set when fizmo opens the interface (and when the
   // values are used in the current version). They keep the current state of
@@ -113,6 +119,8 @@ typedef struct
   bool found_end_of_buffer;
   bool rewound_paragraph_was_newline_terminated;
   bool first_iteration_done;
+  bool validation_disabled;
+  bool dont_skip_newline;
 
   bool metadata_at_index_evaluated;
   z_font font_at_index;
@@ -146,6 +154,7 @@ typedef struct
   z_style saved_last_used_metadata_state_style;
   z_colour saved_last_used_metadata_state_foreground;
   z_colour saved_last_used_metadata_state_background;
+  z_ucs *last_paragraph_attribute_index;
 } history_output;
 
 
@@ -166,9 +175,13 @@ int store_metadata_in_history(OUTPUTHISTORY *history, int metadata_type, ...);
 void store_data_in_history(OUTPUTHISTORY *h, z_ucs *data, size_t len,
     bool evaluate_state_block);
 int remove_chars_from_history(OUTPUTHISTORY *history, int nof_chars);
-z_ucs *get_current_line(OUTPUTHISTORY *history);
-history_output *init_history_output(OUTPUTHISTORY *h, history_output_target *t);
-int output_rewind_paragraph(history_output *output, long *char_count);
+//z_ucs *get_current_line(OUTPUTHISTORY *history);
+history_output *init_history_output(OUTPUTHISTORY *h, history_output_target *t,
+    int output_init_flags);
+int output_rewind_paragraph(history_output *output, long *char_count,
+    int *paragraph_attr1, int *paragraph_attr2);
+int alter_last_paragraph_attributes(history_output *output,
+    int paragraph_attr1, int paragraph_attr2);
 int output_repeat_paragraphs(history_output *output, int n,
     bool include_metadata, bool advance_history_pointer);
 void destroy_history_output(history_output *output);
