@@ -3,7 +3,7 @@
  *
  * This file is part of fizmo.
  *
- * Copyright (c) 2009-2014 Christoph Ender.
+ * Copyright (c) 2009-2015 Christoph Ender.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -888,11 +888,12 @@ int save_game_to_stream(uint16_t address, uint16_t length, z_file *save_file,
         (outputhistory[0]->z_history_buffer_size > 0)
        )
     {
-      history = init_history_output(outputhistory[0], NULL);
+      history = init_history_output(
+          outputhistory[0], NULL, Z_HISTORY_OUTPUT_WITHOUT_EXTRAS);
 
       do
       {
-        return_code = output_rewind_paragraph(history, NULL);
+        return_code = output_rewind_paragraph(history, NULL, NULL, NULL);
         nof_paragraphs_to_save--;
       }
       while ( (nof_paragraphs_to_save > 0) && (return_code == 0) );
